/*
 * Copyright (c) 2010 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */



#if !defined(AFX_CALC_CHECKSUM__H__INCLUDED_)
#define AFX_CALC_CHECKSUM__H__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000





__inline unsigned short FastCalcChecksum(unsigned char *pBuff,size_t dwSize)
{
	unsigned char *pBuffMax;
	unsigned long dwChksum;

	dwChksum=0;
	pBuffMax=(pBuff+(dwSize&~3));
	for(;pBuffMax>pBuff;pBuff+=sizeof(unsigned long))
	{
		dwChksum+=((*((unsigned short*)pBuff))+(*(((unsigned short*)pBuff)+1)));
		//if (dwChksum&0x80000000) dwChksum=((dwChksum&0xFFFF)+(dwChksum>>16));
	}
	if (dwSize&2) {dwChksum+=(*((unsigned short*)pBuff));pBuff+=sizeof(unsigned short);}
	if (dwSize&1) dwChksum+=*pBuff;

	//Complete the calculations
	while(dwChksum>>16) dwChksum=((dwChksum&0xFFFF)+(dwChksum>>16));
  
return((unsigned short)(~dwChksum));
}




__inline unsigned short FastCalcIPv4Checksum(unsigned char *pBuff)
{
	unsigned long dwChksum;
	unsigned short *pusBuff;

	pusBuff=(unsigned short*)pBuff;
	dwChksum=pusBuff[0]+pusBuff[1]+pusBuff[2]+pusBuff[3]+pusBuff[4]+pusBuff[5]+pusBuff[6]+pusBuff[7]+pusBuff[8]+pusBuff[9];

	//Complete the calculations
	while(dwChksum>>16) dwChksum=((dwChksum&0xFFFF)+(dwChksum>>16));

return((unsigned short)(~dwChksum));
}


__inline unsigned short FastCalcUDPv4Checksum(unsigned long dwSrcIP,unsigned long dwDstIP,unsigned char *pBuff,size_t dwSize)
{
	unsigned char *pBuffMax;
	unsigned long dwChksum;

	// add the UDP pseudo header which contains the IP source and destinationn addresses
	// the protocol number and the length of the UDP packet
	// 4352 = ntohs(17), 17 = ip proto udp
	dwChksum=((unsigned short)dwSrcIP)+(*(((unsigned short*)&dwSrcIP)+1))+((unsigned short)dwDstIP)+(*(((unsigned short*)&dwDstIP)+1))+4352+NTOHS(dwSize);

	pBuffMax=(pBuff+(dwSize&~3));
	for(;pBuffMax>pBuff;pBuff+=sizeof(unsigned long))
	{
		dwChksum+=((*((unsigned short*)pBuff))+(*(((unsigned short*)pBuff)+1)));
		//if (dwChksum&0x80000000) dwChksum=((dwChksum&0xFFFF)+(dwChksum>>16));
	}
	if (dwSize&2) {dwChksum+=(*((unsigned short*)pBuff));pBuff+=sizeof(unsigned short);}
	if (dwSize&1) dwChksum+=*pBuff;

	//Complete the calculations
	while(dwChksum>>16) dwChksum=((dwChksum&0xFFFF)+(dwChksum>>16));
  
return((unsigned short)(~dwChksum));
}







#endif // !defined(AFX_CALC_CHECKSUM__H__INCLUDED_)

