/*
 * Copyright (c) 2010 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/cpuset.h>
#include <sys/stat.h>
#include <errno.h>
#include <iostream>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>



#include <default.h>
#include <uTP.h>
#include <CalcChecksum.h>


#define THREAD_BUFF_SIZE	(256*1024)




struct PROG_SETTINGS
{
	int				iSktDivert;
	unsigned short	usBindPort;
	char			*pPidFileName;
	int				iCPUID;
	uid_t			pwUID;// user uid
	gid_t			pwGID;// user gid
};


PROG_SETTINGS gpsProgSettings;




int					GetOptions				(int argc, char **argv);
void				SigInstall				();
static void			SigHandler				(int iSigNum);
void				MakeDaemon				();
int					WritePID				(char *pPIDFileName);
int					BindThreadToCPU			(int iCPUID);
int					SetUserAndGroup			(uid_t pwUID,gid_t pwGID);
int					ProcessPacketLoop		(int iSocket);




int main(int argc,char *argv[])
{
	int iRet=0;
	
	bzero(&gpsProgSettings,sizeof(gpsProgSettings));

	gpsProgSettings.iCPUID=-1;
	if (GetOptions(argc,argv)==0)
	{
		fprintf(stdout,"%s %s: started\n",PACKAGE,VERSION);

		gpsProgSettings.iSktDivert=socket(PF_INET,SOCK_RAW,IPPROTO_DIVERT);
		if (gpsProgSettings.iSktDivert!=-1)
		{// raw divert socket created
			struct sockaddr_in siAddr;

			siAddr.sin_family=PF_INET;
			siAddr.sin_port=HTONS(gpsProgSettings.usBindPort);
			siAddr.sin_addr.s_addr=INADDR_ANY;

			iRet=bind(gpsProgSettings.iSktDivert,(struct sockaddr*)&siAddr,sizeof(siAddr));
			if (iRet!=-1) 
			{// socket binded to specified port
				SigInstall();

				BindThreadToCPU(gpsProgSettings.iCPUID);
				WritePID(gpsProgSettings.pPidFileName);// Store pid to file
				SetUserAndGroup(gpsProgSettings.pwUID,gpsProgSettings.pwGID);// drop rights

				// receive and process packets
				ProcessPacketLoop(gpsProgSettings.iSktDivert);


				if (gpsProgSettings.pPidFileName) unlink(gpsProgSettings.pPidFileName);// Remove pid file
			}else{
				fprintf(stderr,"%s: cant bind() divert socket: %i %s\n",PACKAGE,errno,strerror(errno));
			}
			close(gpsProgSettings.iSktDivert);
		}else{
			fprintf(stderr,"%s: socket() cant create divert socket: %i %s\n",PACKAGE,errno,strerror(errno));
		}

		fprintf(stdout,"%s: exiting\n",argv[0]);
	}else{
		fprintf(stderr, "%s %s -- (c) Rozhuk Ivan <rozhuk.im@gmail.com>\n",PACKAGE,VERSION);
		fprintf(stderr, "   BSD licence.  Website: http://rozhuk.org.ru\n");
		fprintf(stderr,
		       "usage: %s [-d] [-b port] [-c CPUnum] [-p port]\n"
                       "       [-p PID file] [-u uid|usr -g gid|grp]\n"
		       " -h           usage (this screen)\n"
		       " -d           become daemon\n"
		       " -b port      bind to port\n",PACKAGE);
		fprintf(stderr,
		       " -c CPUnum    bind thread to CPU (default no bind)\n"
		       " -p PID file  file name to store PID\n"
		       " -u uid|user  change uid\n"
		       " -g gid|group change gid\n");
	}
return(iRet);
}


int GetOptions(int argc, char **argv)
{
	int iRet=0;
	
	if (argc>1)
	{
		unsigned char *pCur;

		iRet=0;
		for (int i=1;i<argc;i++)
		{
			pCur=(unsigned char*)argv[i];

			if (*pCur++!='-')
			{
				fprintf(stderr,"invalid option: \"%s\"\n", argv[i]);
				iRet=1;
				break;
			}

			while(*pCur && iRet==0)
			{
				switch(*pCur++){
				case '?':case 'h':case 'H':
					iRet=1;
					break;
				case 'd':case 'D':
					MakeDaemon();
					break;
				case 'b':case 'B':
					if (*pCur)
					{
						gpsProgSettings.usBindPort=atol((char*)pCur);
					}else
					if (argv[++i])
					{
						gpsProgSettings.usBindPort=atol(argv[i]);
					}else{
						fprintf(stderr,"option \"-b\" requires bind port number\n");
						iRet=1;
					}
					break;
				case 'c':case 'C':
					if (*pCur)
					{
						gpsProgSettings.iCPUID=atol((char*)pCur);
					}else
					if (argv[++i])
					{
						gpsProgSettings.iCPUID=atol(argv[i]);
					}else{
						fprintf(stderr,"option \"-c\" requires CPU number\n");
						iRet=1;
					}
					break;
				case 'p':case 'P':
					if (*pCur)
					{
						gpsProgSettings.pPidFileName=(char*)pCur;
					}else
					if (argv[++i])
					{
						gpsProgSettings.pPidFileName=argv[i];
					}else{
						fprintf(stderr,"option \"-p\" requires pid file name\n");
						iRet=1;
					}
					break;
				case 'u':case 'U':
					{
						char *pUID;

						if (*pCur)
						{
							pUID=(char*)pCur;
						}else
						if (argv[++i])
						{
							pUID=argv[i];
						}else{
							fprintf(stderr,"option \"-u\" requires UID\n");
							pUID=NULL;
							iRet=1;
						}

						if (pUID)
						{
							struct passwd *pwd;

							pwd=getpwnam(pUID);
							if (pwd)
							{
								gpsProgSettings.pwUID=pwd->pw_uid;
							}else{
								fprintf(stderr,"option \"-u\" requires UID, UID %s not found\n",pUID);
							}
						}
					}
					break;
				case 'g':case 'G':
					{
						char *pGID;

						if (*pCur)
						{
							pGID=(char*)pCur;
						}else
						if (argv[++i])
						{
							pGID=argv[i];
						}else{
							fprintf(stderr,"option \"-g\" requires GID\n");
							pGID=NULL;
							iRet=1;
						}

						if (pGID)
						{
							struct group *grp;

							grp=getgrnam(pGID);
							if (grp)
							{
								gpsProgSettings.pwGID=grp->gr_gid;
							}else{
								fprintf(stderr,"option \"-g\" requires GID, GID %s not found\n",pGID);
							}
						}
					}
					break;
				default:
					fprintf(stderr,"invalid option: \"%c\"\n",*(pCur-1));
					iRet=1;
				}
			}
		}
	}else{
		iRet=1;
	}
return(iRet);
}


void SigInstall()
{
	signal(SIGINT,	SigHandler);
	signal(SIGTERM,	SigHandler);
	signal(SIGKILL,	SigHandler);
	signal(SIGHUP,	SigHandler);
	signal(SIGUSR1,	SigHandler);
	signal(SIGUSR2,	SigHandler);
	signal(SIGPIPE,	SIG_IGN);
}

static void SigHandler(int iSigNum)
{
	switch(iSigNum){
	case SIGINT:
	case SIGTERM:
	case SIGKILL:
		close(gpsProgSettings.iSktDivert);
		break;
	case SIGHUP:
	case SIGUSR1:
	case SIGUSR2:
		break;
	}
}


void MakeDaemon()
{
    pid_t pdDaemon;

	pdDaemon=fork();
    if (pdDaemon!=-1)
	{
		if (pdDaemon)
		{//Exit Parent
			exit(0);
		}else{//Child
			setsid();
			setpgid(getpid(),0);
			chdir("/tmp/");
			umask(077);

			//Close stdin/stdout/stderr
			close(0);
			close(1);
			close(2);
		}
	}else{
		fprintf(stderr,"MakeDaemon: fork() failed: %i %s\n",errno,strerror(errno));
	}
}


int WritePID(char *pPIDFileName)
{
	int iRet=0;

	if (pPIDFileName)
	{
		int iFileHandle;
		
		iFileHandle=open(pPIDFileName,(O_RDWR|O_CREAT|O_TRUNC));
		if (iFileHandle==-1)
		{
			fprintf(stderr,"WritePID: fopen() failed: %i %s\n",errno,strerror(errno));
		}else{
			char szPid[64];
			size_t dwPidSize;
			
			dwPidSize=snprintf(szPid,sizeof(szPid),"%ld",getpid());
			write(iFileHandle,szPid,dwPidSize);
			fchmod(iFileHandle,(S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH));
			close(iFileHandle);
			iRet=1;
		}
	}
return(iRet);
}



int BindThreadToCPU(int iCPUID)
{
	int bRet=0;

	if (iCPUID!=-1)
	{
		cpuset_t csCPUSet;

		// bind this thread to a single cpu
		CPU_ZERO(&csCPUSet);
		CPU_SET(iCPUID,&csCPUSet);
		if (cpuset_setaffinity(CPU_LEVEL_WHICH,CPU_WHICH_TID,-1,sizeof(csCPUSet),&csCPUSet)==-1)
		{
			fprintf(stderr,"BindThreadToCPU: can't bind thread to CPU %s\n",strerror(errno));
		}else{
			bRet=1;
		}
	}
return(bRet);
}


int SetUserAndGroup(uid_t pwUID,gid_t pwGID)
{
	int bRet=0;


	if (pwUID && pwGID)
	{
		struct passwd *pwd;

		pwd=getpwuid(pwUID);
		if (pwd)
		{
			if (setgid(pwGID)!=0) fprintf(stderr,"setgid() failed: %i %s\n",errno,strerror(errno));
#ifdef HAVE_INITGROUPS
			if (initgroups(pwd->pw_name,pwGID)!=0) fprintf(stderr,"initgroups() failed: %i %s\n",errno,strerror(errno));
#else
#if HAVE_SETGROUPS
			if (setgroups(1,&pwd->pw_gid)!=0) fprintf(stderr,"initgroups() failed: %i %s\n",errno,strerror(errno));
#endif
#endif
			if (setuid(pwUID)!=0) fprintf(stderr,"setuid() failed: %i %s\n",errno,strerror(errno));

			bRet=1;
		}else{
			fprintf(stderr,"getpwuid() failed: %i %s\n",errno,strerror(errno));
		}
	}
return(bRet);
}






int ProcessPacketLoop(int iSocket)
{
	int iRet=0;


	int iVersion;
	unsigned int uiConnIDSend,uiFlagsType;
	unsigned short usAckNr,usSeqNr;
	size_t dwDataSize,dwUDPDataSize;
	socklen_t slIFAddrLen;
	struct ip *pIPHdr;
	struct udphdr *pUDPHdr;
	sockaddr_in siIFAddr;
	unsigned char *pBuff,*pUDPData,btBuff[THREAD_BUFF_SIZE];


	pBuff=(unsigned char*)btBuff;
	pIPHdr=(struct ip*)pBuff;
	pUDPHdr=(struct udphdr*)(pIPHdr+1);
	pUDPData=(unsigned char*)(pUDPHdr+1);

	// read data in
	while(1)
	{
		slIFAddrLen=sizeof(siIFAddr);
		dwDataSize=recvfrom(iSocket,pBuff,THREAD_BUFF_SIZE,0,(struct sockaddr*)&siIFAddr,&slIFAddrLen);
		if (dwDataSize!=(size_t)-1)
		{// received some packet
			if (((struct ip*)pBuff)->ip_p==IPPROTO_UDP)// siIFAddr.sin_addr.s_addr!=INADDR_ANY && 
			{
				dwUDPDataSize=(NTOHS(pUDPHdr->uh_ulen)-sizeof(struct udphdr));

				/*{// check summ verify
					unsigned short usIPcsOld,usIPcs1,usUDPcsOld,usUDPcs1;

					printf("\n\n");

					//bzero((pBuff+dwDataSize),(sizeof(btBuff)-dwDataSize));
					memset((pBuff+dwDataSize),1,128);
					usIPcsOld=pIPHdr->ip_sum;
					pIPHdr->ip_sum=0;
					usIPcs1=FastCalcIPChecksum((unsigned char*)pIPHdr);//CalculateChecksum((unsigned char*)pIPHdr,sizeof(struct ip));
					pIPHdr->ip_sum=usIPcsOld;

					usUDPcsOld=pUDPHdr->uh_sum;
					pUDPHdr->uh_sum=0;
					usUDPcs1=FastCalcUDPChecksum(pIPHdr->ip_src.s_addr,pIPHdr->ip_dst.s_addr,(unsigned char*)pUDPHdr,NTOHS(pUDPHdr->uh_ulen));
					pUDPHdr->uh_sum=usUDPcsOld;

					printf("\nIP checksum=%hX, UDP checksum=%hX, len=%hu or =%lu\n",usIPcsOld,usUDPcsOld,NTOHS(pUDPHdr->uh_ulen),(dwDataSize-(((unsigned char*)pUDPHdr)-pBuff)));
					printf("IP checksum=%hX, UDP checksum=%hX\n",usIPcs1,usUDPcs1);

					if (usIPcsOld!=usIPcs1 || (usUDPcsOld && usUDPcsOld!=usUDPcs1)) exit(0);
				}//*/
				
				iVersion=GetPacketuTPVersion(pUDPData,dwUDPDataSize);
				if (iVersion!=-1)
				{
					uiFlagsType=GetFlagsTypeFromPacketuTP(iVersion,pUDPData);
					
					/*{// detailed info per detected packet
						char szIfName[16];
						time_t tv_usec=GetTVUsecFromPacketuTP(iVersion,pUDPData);//,ctime(&tv_usec)

						memmove(szIfName,(((unsigned char*)(&siIFAddr.sin_addr))+4),8);
						szIfName[9]=0;

						printf("\nAddress: %s",inet_ntoa(pIPHdr->ip_src));
						printf(" -> %s",inet_ntoa(pIPHdr->ip_dst));
						printf(" via %s, ipfw rule=%hhu, IF name=%s, Protocol number=%hhu, checksum=%hu, ip_len=%hu, BuffSize=%lu\n",inet_ntoa(siIFAddr.sin_addr),NTOHS(siIFAddr.sin_port),szIfName,pIPHdr->ip_p,NTOHS(pIPHdr->ip_sum),NTOHS(pIPHdr->ip_len),dwDataSize);
						printf("UDP from port %hu to %hu, checksum=%hu, len=%hu, UDPDataSize=%lu or =%lu\n",HTONS(pUDPHdr->uh_sport),HTONS(pUDPHdr->uh_dport),NTOHS(pUDPHdr->uh_sum),NTOHS(pUDPHdr->uh_ulen),dwUDPDataSize,(dwDataSize-(pUDPData-pBuff)));
						printf("uTP Version=%i, TypeFlag=%s, ext=%hhu, conn_id_send=%lu, ack_nr=%hu, seq_nr=%hu, windowsize=%lu, reply_micro=%lu, tv_usec=%lu\n",iVersion,flagnames[uiFlagsType],GetExtFromPacketuTP(iVersion,pUDPData),GetConnIDFromPacketuTP(iVersion,pUDPData),GetAckNrFromPacketuTP(iVersion,pUDPData),GetSeqNrFromPacketuTP(iVersion,pUDPData),GetWindowSizeFromPacketuTP(iVersion,pUDPData),GetReplyMicroFromPacketuTP(iVersion,pUDPData),tv_usec);
					}//*/

					//if (uiFlagsType==ST_STATE || uiFlagsType==ST_SYN)
					if (uiFlagsType!=ST_FIN && uiFlagsType!=ST_RESET)
					{// send reset with all packet types, except reset, fin
						uiConnIDSend=GetConnIDFromPacketuTP(iVersion,pUDPData);
						usAckNr=GetAckNrFromPacketuTP(iVersion,pUDPData);
						if (iVersion)
						{
							usSeqNr=(usAckNr+1);
						}else{
							usSeqNr=GetSeqNrFromPacketuTP(iVersion,pUDPData);
						}

						{// generate reply! (instead duplicate packet and make them rst)
							dwDataSize=pIPHdr->ip_src.s_addr;
							pIPHdr->ip_src.s_addr=pIPHdr->ip_dst.s_addr;
							pIPHdr->ip_dst.s_addr=dwDataSize;

							usAckNr=pUDPHdr->uh_sport;
							pUDPHdr->uh_sport=pUDPHdr->uh_dport;
							pUDPHdr->uh_dport=usAckNr;

							usAckNr=GetSeqNrFromPacketuTP(iVersion,pUDPData);
							usSeqNr=(usAckNr+1);
							siIFAddr.sin_addr.s_addr=INADDR_ANY;
						}//*/

						dwUDPDataSize=GenRSTPacketuTP(iVersion,uiConnIDSend,usAckNr,usSeqNr,pUDPData,(THREAD_BUFF_SIZE-(sizeof(struct ip)+sizeof(struct udphdr))));
						if (dwUDPDataSize)
						{
							dwUDPDataSize+=sizeof(struct udphdr);
							pUDPHdr->uh_ulen=HTONS(dwUDPDataSize);
							pUDPHdr->uh_sum=0;// auto recalc sum UDP not work on some IF types, like ng
							pUDPHdr->uh_sum=FastCalcUDPv4Checksum(pIPHdr->ip_src.s_addr,pIPHdr->ip_dst.s_addr,(unsigned char*)pUDPHdr,dwUDPDataSize);
							pIPHdr->ip_len=HTONS((dwUDPDataSize+sizeof(struct ip)));
							pIPHdr->ip_sum=0;// auto recalc sum IP not work on some IF types, like ng
							pIPHdr->ip_sum=FastCalcIPv4Checksum((unsigned char*)pIPHdr);

							iRet=sendto(iSocket,pBuff,(dwUDPDataSize+sizeof(struct ip)),MSG_NOSIGNAL,(struct sockaddr*)&siIFAddr,slIFAddrLen);
							if (iRet==-1)
							{
								fprintf(stderr,"ProcessPacket: sendto() error %i %s\n",errno,strerror(errno));
							}
						}
					}
				}
			}
		}else{// err on recvfrom, exit on error
			fprintf(stderr,"%s: recvfrom error: %i %s\n",PACKAGE,errno,strerror(errno));
			break;
		}
	}// end while

return(iRet);
}



